% :- type feladvany_leiro ---> csiga(meret, ciklus, list(adott_elem)).
% :- type meret             == integer.
% :- type ciklus            == integer.
% :- type adott_elem      ---> i(sorszam, oszlopszam, ertek).
% :- type sorszam           == integer.
% :- type oszlopszam        == integer.
% :- type ertek             == integer.
		          
% :- type csigatabla        == list(sor).
% :- type sor               == list(ertek_vagy_ures).
% :- type ertek_vagy_ures   == integer.

% :- type file == atom.

% :- pred buvos_csiga(feladvany_leiro::in, csigatabla::out).

:- module(csiga_keret,
	  [
	   csiga_be/2,		% csiga_be(file::in, feladvany_leiro::out).
	   csiga_ki/2,		% csiga_ki(file::in, csigatabla::in).
	   megold/2,		% megold(file::in, file::in)
	   stopper/2		% stopper(file::in, file::in)
	  ]
	 ).

:- use_module(library(lists), [select/3,member/2,append/3,
			       max_list/2,nextto/3]).

%-------------- Bejaratok ------------------------------------

% csiga_be(+File, -Leiro):
% A File-bol beolvashato egy feladat, melyet a Leiro ir le.
csiga_be(File, Leiro) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
		     close(Stream)),
	parse_data(Characters, Leiro).

% csiga_ki(+File, +Tabla):
% A Tabla megoldast kiirja a File allomanyba.
csiga_ki(File, Tabla) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Tabla),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	csiga_be(FileIn, Leiro),
	ensure_loaded(user:csiga),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Leiro),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Problem-file: ~a, time: ~3d s~n', [FileIn,T]).

%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.  Ha C = -1,
% akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Descr):
% Chars fuzerbol kielemezheto egy a Descr leiroval megadott feladat
parse_data(Chars, Descr) :-
	descriptor(Descr, Chars, Rest), !,
	correct_syntax(Rest),
	correct_descr(Descr).
parse_data(_, _) :-
	warning('Bad format of input file.',[]),
	fail.

%-------------- Ellenorzesek ---------------------------------

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !, 
	chars_of_type(mid_line, L, Rest, _),
	warning('unexpected character in line: "~s"', [L]),
	fail.
correct_syntax(_).

% correct_descr(+Descr):
% A Descr altal megadott feladat lehetseges.
correct_descr(csiga(Size,Cycle,RCVs)) :-
	integer(Size), integer(Cycle),
	Size >= 1, Cycle >= 1,
	Cycle =< Size,
        % nincs nem helyes eleme az RCVs listnak:
	\+ (   member(RCV, RCVs), 
	       \+ correct_elem(Size, Cycle, RCV)
	   ),  
	% az RCVs lista lexikografikusan rendezett:
	sort(RCVs, RCVs),
	% Az RCVs lista minden eleme klnbz (Sor,Oszlop) 
	% koordintra vonatkozik (a rendezettsg miatt elegend a
        % szomszdos elemekre kiktni ezt):
	\+ nextto(i(Row,Col,_), i(Row,Col,_), RCVs),
	!.
correct_descr(_Descr) :-
	warning('incorrect data', []), fail.

correct_elem(Size, Cycle, i(Row,Col,Val)) :-
	integer(Row), integer(Col), integer(Val),
	1 =< Row, Row =< Size,
	1 =< Col, Col =< Size,
	1 =< Val, Val =< Cycle.

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy feladvany-leiro
descriptor(csiga(Size,Cycle,RCVs)) -->
	nat_number(Size),
	newline,
	nat_number(Cycle),
	newline,
	triplet_list(RCVs),
	chars_of_type(layout, _).

% Kielemezheto egy (tizes szamrendszerbeli) Int termeszetes szam, amit
% esetleg szokozok eloznek meg.
nat_number(Int) -->
	chars_of_type(horizontal, _),
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% triplet_list(RCVs)
% Kielemezheto egy harmasokbol allo RCVs lista.
triplet_list([i(R,C,V)|RCVs]) -->
	nat_number(R), nat_number(C), nat_number(V),
	newline, !,
	triplet_list(RCVs).
triplet_list([]) --> [].

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).



% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(   vertical_layout_char(C) -> true
	;   horizontal_layout_char(C)
	).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).

%-------------- Kiiratas -------------------------------------

% print_solutions(+Descr):
% A Descr altal megadott feladat osszes megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Descr) :-
	user:buvos_csiga(Descr, Table),
	pretty_print(Table),
	fail.
print_solutions(_).

% pretty_print(+Solution, +Descr):
% kiirja a Solution megoldast, amely a Descr leiroju feladathoz tartozik.
pretty_print(Solution) :-
	Solution = [FRow|_],
	length(FRow, NumCols),
	matrix_elements_max(Solution, _, MaxNum),
	Pad is integer(log(10, max(MaxNum,1)))+1,
	pretty_print_rows(Solution, 1, Pad, NumCols).

% pretty_print_rows(+Ss, +RNo +Pad, +NumCols):
% Az RNo sorban kezdodo Ss megoldassorokat kiirja, a szamokat
% egy Pad szelessegu mezoben jobbra igazitva.
pretty_print_rows([R|Rs], RNo, Pad, NumCols) :-
	(   RNo = 1 ->
	    pretty_print_underline(R, 0, 1, Pad, NumCols),
	    put_code(0' )
	;   put_code(0'|)
	),
	pretty_print_row(R, RNo, 1, Pad, NumCols),
	corner_char(RNo, 0, NumCols, CornerC),
	put_code(CornerC),
	pretty_print_underline(R, RNo, 1, Pad, NumCols),
	RNo1 is RNo + 1,
	pretty_print_rows(Rs, RNo1, Pad, NumCols).
pretty_print_rows([], _, _, _).


% pretty_print_row(+R, +RNo, +CNo, +Pad, +NumCols):
% Az (RNo,CNo) helyen kezdodo R megoldassort kiirja
% (a szamokat egy min. Pad szelessegu mezoben jobbra igazitva) 
pretty_print_row([], _, _, _, _) :-
	nl.
pretty_print_row([V0|Vs], RNo, CNo, Pad, NumCols) :-
	right_char(RNo, CNo, NumCols, RightC),
	value_term(V0, V),
	format(' ~|~t~w~*+ ~c', [V,Pad,RightC]),
	CNo1 is CNo+1,
	pretty_print_row(Vs, RNo, CNo1, Pad, NumCols).

% pretty_print_underline(+R, +RNo, +CNo, +Pad, +NumCols):
% Kiirja az (RNo,CNo) helyen kezdodo R sorhoz tartozo alahuzas-sort,
% feltetelezve hogy egy szam hossza Pad.
pretty_print_underline([], _, _, _, _) :-
	nl.
pretty_print_underline([_|R], RNo, CNo, Pad, NumCols) :-
	bottom_char(RNo, CNo, NumCols, BottomC),
	corner_char(RNo, CNo, NumCols, CornerC),
	N is Pad+2,
	format('~*c~c', [N, BottomC,CornerC]),
	CNo1 is CNo+1,
	pretty_print_underline(R, RNo, CNo1, Pad, NumCols).

value_term(V0, V) :- var(V0), !, V = ' '.
value_term(0, V) :- !, V = '-'.
value_term(V, V).

% right_char(+RNo, +CNo, +NumCols, ?RightC):
% Az (RNo,CNo) sor-oszlopszamu mezo jobb hatarolo karaktere RightC.
right_char(RNo, CNo, NumCols, RightC) :-
	(   RNo+CNo >= NumCols+1 -> CNo-RNo >= 0  % keleti negyed 
	;   CNo-RNo < -1                          % nyugati negyed
	), !, RightC = 0'|.
right_char(_RNo, _CNo, _NumCols, 0' ).

% bottom_char(+RNo, +CNo, +NumCols, ?BottomC)
% Az (RNo,CNo) sor-oszlopszamu mezo also hatarolo karaktere BottomC.
bottom_char(RNo, CNo, NumCols, BottomC) :-
	(   RNo+CNo >= NumCols+1 -> CNo-RNo > 0
	;   CNo-RNo =< -1
	), !, BottomC = 0' .
bottom_char(_RNo, _CNo, _NumCols, 0'-).

% corner_char(+RNo, +CNo, +NumCols, ?CornerC) :-
% Az (RNo,CNo) sor-oszlopszamu mezo sarok-hatarolo karaktere CornerC.
corner_char(RNo, CNo, NumCols, CornerC) :-
	% ha a mezo also hataroloja a `-' karakter:
	bottom_char(RNo, CNo, NumCols, 0'-),
	% es a jobb szomszed mezo also hataroloja is a `-' karakter:
	CNo1 is CNo+1,
	bottom_char(RNo, CNo1, NumCols, 0'-),
	!,
	% akkor a sarok karakter a `-' legyen.
	CornerC = 0'-.
corner_char(RNo, CNo, NumCols, CornerC) :-
	% A DNy-EK atloban a sarok karakter a `\'
	RNo+CNo =:= NumCols, !, CornerC = 0'\\.
corner_char(RNo, CNo, _NumCols, CornerC) :-
	% A ENy-DK atloban es az alatta levo atlos sorban
	% a sarok karakter a `/', ha a korabbi feltetelek nem 
	% teljesultek.
	(   RNo =:= CNo+1 -> true
	;   RNo =:= CNo
	),
	!, CornerC = 0'/.
corner_char(_RNo, _CNo, _NumCols, 0'|).

%-------------- Seged-eljarasok ------------------------------

% matrix_elements_max(+LLs, ?All, ?Max):
% Az LLs matrix osszes elemenek listaja All, ezek maximuma Max.
matrix_elements_max(LLs, All, Max) :-
	append_lists(LLs, All),
	max_list(All, Max).

% append_lists(+Ls, ?L):
% Az Ls listak osszefuzesevel kapott lista L.
append_lists([], []).
append_lists([H|T], C) :-
	append(H, R, C),
	append_lists(T, R).

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
